/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <algorithm>

using namespace std;

# include "AssignAtomIDs.h"
# include "AtomCode.h"
# include "AtomUtils.h"
# include "error.h"
# include "GetAtomType.h"
# include "GetAtomType2.h"
# include "ReadPdb.h"
# include "ResidueCode.h"
# include "ResidueID.h"
# include "Structures.h"
# include "StringUtils.h"
# include "TypeDef.h"
# include "LinkedList.h"
# include "IOUtils.h"
# include "VectorManip.h"

int CountAtomsInPdb(string PdbFile)
{
	bool valid;
	//bool verbose=false;
	char CharPdbFile[1000];
	string line, ElementID, StrResidueID, StrAtomName;
	int natom;
	natom=0;
	strcpy(CharPdbFile, PdbFile.c_str());
	valid=exist(CharPdbFile);
	if (!valid)
	{
		cout <<"Error could not open pdb file "<<CharPdbFile<<endl;
		exit(EXIT_FAILURE);
	}

	fstream protein;
	protein.open(CharPdbFile, ios::in);

	getline(protein,line);

	while (true)
	{

		if (protein.eof()) break;
		if (line.substr(0,3)=="END") break;
		//if (line.substr(0,3)=="TER") break;


		if (line.substr(0,4)=="ATOM" || line.substr(0,4)=="HETA")
		{
			natom++;
		}
		getline(protein,line);
	}
	protein.close();
	//cout <<"natom= "<<natom<<endl;
	return natom;
}

void readPdb(string PdbFile, ProteinStruct &Protein)
{
	AtomStruct Atom;
	//bool verbose=false;
	string line;
	vector<string> str;
	int coord, currentIndex=0, natom, n, Size;
	SymmetryOperationStruct symmetryOperation;
	fstream file;
	
	DeleteVector(Protein.Atoms);
	OpenFile(PdbFile, file, "pdb file");
	natom=CountAtomsInPdb(PdbFile);
	Protein.symmetryOperations.clear();
	SafeAlloc(Protein.Atoms, Atom, natom, "Atoms in readPdb");
	SafeAlloc(symmetryOperation.translationVector, 3, "translationVector");
	Safe2DAlloc(symmetryOperation.rotationMat, 3, 3, "rotationMat");
	Protein.name=PdbFile;
	getline(file, line);

	while (true)
	{
		line=line+"                                         ";

		if (file.eof()) break;
		if (line.substr(0,4)=="CRYS")
		{
			Tokenize2(line, " ", str);
			Size=str.size();
			if (Size>=7)
			{
				Protein.XBoxLength=toReal(str[1]);
				Protein.YBoxLength=toReal(str[2]);
				Protein.ZBoxLength=toReal(str[3]);
				Protein.alpha=toReal(str[4]);
				Protein.beta=toReal(str[5]);
				Protein.gamma=toReal(str[6]);
			}
			if (Size>=8) Protein.bravisLatticeCentering=str[7];
			Protein.spaceGroup=str[7];
			if (Size>9) 
			{
				Protein.symmetry1=toInt(str[8]);
				Protein.spaceGroup+=str[8];
			}
			if (Size>10) 
			{
				Protein.symmetry2=toInt(str[9]);
				Protein.spaceGroup+=str[9];
			}
			if (Size>11) 
			{
				Protein.symmetry3=toInt(str[10]);
				Protein.spaceGroup+=str[10];
			}
			if (Size>12) 
			{
				Protein.spaceGroup+=str[11];
			}
			//Protein.numPerUnitCell=StrToInt(GetWord(line, 12));
		}
		if (GetWord(line,1)=="REMARK" && GetWord(line,2)=="290")
		{
			if (line.substr(13,5)=="SMTRY")
			{
				coord=StrToInt(line.substr(18,1))-1;
				symmetryOperation.rotationMat[coord][X]=StrToFloat(GetWord(line,5));
				symmetryOperation.rotationMat[coord][Y]=StrToFloat(GetWord(line,6));
				symmetryOperation.rotationMat[coord][Z]=StrToFloat(GetWord(line,7));
				symmetryOperation.translationVector[coord]=StrToFloat(GetWord(line,8));
				if (coord==Z)
				{
					SafePushBack(Protein.symmetryOperations, symmetryOperation, "symmetryOperations");
				}
			}
		}
		if (line.substr(0,5)=="SCALE")
		{
			coord=StrToInt(line.substr(5,1))-1;
			if (coord==X) Safe2DAlloc(Protein.cartToFrac, 3, 3, "cartToFrac");
			Protein.cartToFrac[coord][X]=StrToFloat(GetWord(line, 2));
			Protein.cartToFrac[coord][Y]=StrToFloat(GetWord(line, 3));
			Protein.cartToFrac[coord][Z]=StrToFloat(GetWord(line, 4));
		}
		if (line.substr(0,7)=="CONNECT")
		{
			Tokenize2(line, " ", str);
			Size=str.size();
			for (int i=2;i<Size;i++)
			{
				n=toInt(str[1])-1;
				if (n<natom)
				{
					SafePushBack(Protein.Atoms[n].connectivity, toInt(str[i])-1, "connectivity");
					if (toInt(str[i])-1>natom)
					{
						cout <<"toInt(str["<<i<<"])-1="<<toInt(str[i])-1<<" n= "<<n<<endl;
						cout <<line<<endl;
						cout <<PdbFile<<endl;
					}
				}
			}
		}
		if (line.substr(0,3)=="END") break;
		//if (line.substr(0,3)=="TER") break;

		if (line.substr(0,4)=="ATOM" || line.substr(0,4)=="HETA")
		{
			PdbLineToAtom(line, Protein.Atoms[currentIndex]);
			currentIndex++;
		}
		if (line.substr(0,6)=="ANISOU")
		{
			readAnisou(line,Protein.Atoms[currentIndex-1].anisou);
		}
		getline(file, line);
	}
	calcUnitCellVectors(Protein);
	file.close();
}

void readPdbs(string pdbListFile, vector<ProteinStruct> &Proteins)
{
	vector<string> lines;
	int nline;
	ProteinStruct tempProtein;

	readLines(pdbListFile, lines, "pdbListFile");
	nline=lines.size();

	for (int i=0;i<nline;i++)
	{
		readPdb(lines[i], tempProtein);
		SafePushBack(Proteins, tempProtein, "Proteins");
	}
}

bool readAnisou(string line, vector<Real> &anisou)
{
	if (line.substr(0,6)=="ANISOU")
	{
		SafeAlloc(anisou, 6, "anisou");
		for (int i=0;i<6;i++)
		{
			anisou[i]=StrToReal(line.substr(29+7*i,6))/10000.0;
		}
		return true;
	}
	return false;
}

bool pdbLineToInfo(string line, bool &HetAtom, int &AtomNumber, string &AtomName, string &duplicate, string &ResidueName, string &ChainName, int &ResidueNum, Real &x, Real &y, Real &z, Real &Occupancy, Real &BFactor, string &SegID, string &ID)
{
	int length=line.length();
	//cout <<line<<endl;
	if (line.substr(0,4)=="ATOM" || line.substr(0,4)=="HETA")
	{
		//cout <<line<<endl;
		if (line.substr(0,4)=="ATOM") HetAtom=false;
		else if (line.substr(0,4)=="HETA") HetAtom=true;
		AtomNumber=StrToInt(line.substr(7,5));
		AtomName=line.substr(12,4);
		duplicate=line.substr(16,1);
		ResidueName=line.substr(17,4);
		ChainName=line.substr(21,1);
		ResidueNum=StrToInt(line.substr(22,4));
		x=StrToFloat(line.substr(30,8));
		y=StrToFloat(line.substr(38,8));
		z=StrToFloat(line.substr(46,8));
		if (line.substr(56,4)!="    " && length>60)
		{
			Occupancy=StrToFloat(line.substr(56,4));
		}
		else Occupancy=1.0;
		if (line.substr(60,9)!="        " && length>68)
		{
			BFactor=StrToFloat(line.substr(60,9));
		}
		else BFactor=1.0;
		SegID=line.substr(72,4);
		ID=line.substr(73,3);
		return true;
	}
	return false;
}

void PdbLineToAtom(string line, AtomStruct &Atom)
{
	pdbLineToInfo(line, Atom.HetAtom, Atom.AtomNumber, Atom.AtomName, Atom.duplicate, Atom.ResidueName, Atom.ChainName, Atom.ResidueNum, Atom.x, Atom.y, Atom.z, Atom.Occupancy, Atom.BFactor, Atom.SegID, Atom.ID);
	Atom.weight=1.0;
	Trim(Atom.AtomName);
	Trim(Atom.ResidueName);
	Atom.residueid=GetResidueID(Atom.ResidueName);

}

void PdbLinesToAtoms(vector<string> &lines, vector<AtomStruct> &Atoms)
{
	int natom=lines.size();
	AtomStruct Atom;

	SafeAlloc(Atoms, Atom, natom, "in LinesToAtoms");
	for (int i=0;i<natom;i++)
	{
		PdbLineToAtom(lines[i], Atoms[i]);
	}
}

int readPdb(string PdbFile, vector<AtomStruct> &Atoms)
{
	int natom=0;
	ProteinStruct Protein;

	readPdb(PdbFile, Protein);
	Atoms=Protein.Atoms;

	return natom;
}

void MinMaxPdb(string PdbFile, Real &xmin, Real &ymin, Real &zmin, Real &xmax, Real &ymax, Real &zmax)
{
	fstream file;
	bool HetAtom;
	string line, AtomName, duplicate, ResidueName, ChainName, SegID, ID;
	int AtomNumber, ResidueNum;
	Real Occupancy, BFactor;
	Real x, y, z;

	xmin=10e10;
	ymin=10e10;
	zmin=10e10;
	xmax=-10e10;
	ymax=-10e10;
	zmax=-10e10;

	OpenFile(PdbFile, file, "pdb file");
	getline(file, line);

	while (true)
	{
		//cout <<line<<endl;
		if (file.eof()) break;
		pdbLineToInfo(line, HetAtom, AtomNumber, AtomName, duplicate, ResidueName, ChainName, ResidueNum, x, y, z, Occupancy, BFactor, SegID, ID);
		if (line.substr(0,4)=="HETA")
		{
			if (x<xmin) xmin=x;
			if (y<ymin) ymin=y;
			if (z<zmin) zmin=z;
			if (x>xmax) xmax=x;
			if (y>ymax) ymax=y;
			if (z>zmax) zmax=z;
		}
		getline(file, line);
	}
}

void ReadXYZ(string xyz, vector<AtomStruct> &Atoms)
{
	AtomStruct Atom;
	bool valid;
	vector<string> str;
	char CharXYZ[1000];
	string line;
	int n, natom, Size;

	cout <<"In ReadXYZ"<<endl;
	strcpy(CharXYZ, xyz.c_str());
	valid=exist(CharXYZ);
	if (!valid)
	{
		cout <<"Error unable to open xyz file "<<CharXYZ<<endl;
		exit(0);
	}

	int AtomNumber=0;
	int AtomName=3;
	int XPos=12;
	int YPos=29;
	int ZPos=41;
	int AtomType=47;
	int Connect1=53;
	int Connect2=55;
	int Connect3=65;
	int Connect4=71;

	fstream protein;
	protein.open(CharXYZ, ios::in);

	getline(protein, line);
	natom=StrToInt(GetWord2(line, 1));
	getline(protein, line);
	AtomNumber=int(log10(Real(natom)))+1;
	AtomName+=AtomNumber;
	XPos+=AtomNumber;
	YPos+=AtomNumber;
	ZPos+=AtomNumber;
	AtomType+=AtomNumber;
	Connect1+=AtomNumber;
	Connect2+=AtomNumber;
	Connect3+=AtomNumber;
	Connect4+=AtomNumber;
	cout <<"AtomNumber= "<<AtomNumber<<endl;
	for (n=0;n<natom;n++)
	{
		//if (n==25) exit(EXIT_FAILURE);
/*
		cout <<"n= "<<n<<endl;
		length=line.length();
		if (protein.eof()) break;
		Atom.AtomNumber=StrToInt(line.substr(0, AtomNumber));
		cout <<line.substr(0, AtomNumber)<<endl;
		cout <<"AtomNumber= "<<Atom.AtomNumber<<endl;
		Atom.AtomName=line.substr(AtomNumber, XPos-AtomName);
		cout <<line.substr(AtomName, XPos-AtomName)<<endl;
		cout <<"AtomName= "<<Atom.AtomName<<endl;
		Atom.x=StrToFloat(line.substr(AtomName, YPos-XPos));
		cout <<line.substr(AtomName, YPos-XPos)<<endl;
		cout <<"x= "<<Atom.x<<endl;
		Atom.y=StrToFloat(line.substr(XPos, ZPos-XPos));
		cout <<line.substr(XPos, ZPos-XPos)<<endl;
		cout <<"y= "<<Atom.y<<endl;
		Atom.z=StrToFloat(line.substr(ZPos, AtomType-ZPos));
		cout <<line.substr(ZPos, AtomType-ZPos)<<endl;
		cout <<"z= "<<Atom.z<<endl;
		Atom.ParmType=StrToInt(line.substr(AtomType, Connect1-AtomType));
		exit(EXIT_FAILURE);
		Atom.connectivity.clear();
		if (length>=Connect2) Atom.connectivity.push_back(StrToInt(line.substr(Connect1, Connect2-Connect1)));
		if (length>=Connect3) Atom.connectivity.push_back(StrToInt(line.substr(Connect2, Connect3-Connect2)));
		if (length>=Connect4) Atom.connectivity.push_back(StrToInt(line.substr(Connect3, Connect4-Connect3)));
		if (length>=100) Atom.connectivity.push_back(StrToInt(line.substr(Connect4, 100)));
*/
		Tokenize2(line, " ", str);
		Size=str.size();
		if (Size>=6)
		{
			Atom.AtomNumber=toInt(str[0]);
			Atom.AtomName=str[1];
			Atom.x=toReal(str[2]);
			Atom.y=toReal(str[3]);
			Atom.z=toReal(str[4]);
			Atom.ParmType=toInt(str[5]);
			Atom.connectivity.clear();
			if (Size>=7) 
			{
				SafePushBack(Atom.connectivity, toInt(str[6]), "connectivity");
			}
			if (Size>=8) 
			{
				SafePushBack(Atom.connectivity, toInt(str[7]), "connectivity");
			}
			if (Size>=9) 
			{
				SafePushBack(Atom.connectivity, toInt(str[8]), "connectivity");
			}
			if (Size>=10) 
			{
				SafePushBack(Atom.connectivity, toInt(str[9]), "connectivity");
			}
			SafePushBack(Atoms, Atom, "Atoms");
		}
		getline(protein, line);
	}
}

void ReadXYZLine(string line, AtomStruct Atom)
{
	int AtomNumber=6;
	int AtomName=12;
	int XPos=24;
	int YPos=36;
	int ZPos=48;
	int AtomType=54;
	int Connect1=60;
	int Connect2=66;
	int Connect3=72;
	int Connect4=78;

	Atom.AtomNumber=StrToInt(line.substr(0, AtomNumber));
	Atom.AtomName=line.substr(AtomNumber, XPos-AtomName);
	Atom.x=StrToFloat(line.substr(AtomName, YPos-XPos));
	Atom.y=StrToFloat(line.substr(XPos, ZPos-XPos));
	Atom.z=StrToFloat(line.substr(ZPos, AtomType-ZPos));
	Atom.ParmType=StrToInt(line.substr(Connect1, Connect2-Connect1));
	Atom.connectivity.push_back(StrToInt(line.substr(Connect2, Connect3-Connect2)));
	Atom.connectivity.push_back(StrToInt(line.substr(Connect3, Connect4-Connect3)));
	Atom.connectivity.push_back(StrToInt(line.substr(Connect4, 100)));
	Atom.connectivity.push_back(StrToInt(line.substr(Connect4, 100)));
}
