/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef _cluster_included_
#define _cluster_included_

# include "MathUtils.h"
# include "error.h"

bool InVector(vector<int> &v, int index)
{
        int Size=v.size();

        for (int i=0;i<Size;i++)
        {
                if (v[i]==index) return true;
        }
        return false;
}

void AssignCentroidsAtRandom(int Size, int NumClusters, vector<int> &centroids)
{
        int pick;
	int ncentroid;
        centroids.clear();
        while (true)
        {
                pick=randInt(Size-1);
		if (pick>=Size || pick<0)
		{
			cout <<"ERROR: pick= "<<pick<<" Size= "<<Size<<endl;
			exit(EXIT_FAILURE);
		}
                if (!InVector(centroids, pick))
                {
                        SafePushBack(centroids, pick, "AssignCentroidsAtRandom");
			ncentroid=centroids.size();
                        if (ncentroid>=NumClusters) break;
                }
        }
}

int FindBestCluster(vector<int> &centroids, Matrix &DistanceMatrix, int index)
{
        int NumClusters=centroids.size();
        int BestCluster;
        Real min;

        min=DistanceMatrix[index][centroids[0]];
        BestCluster=0;
        for (int i=1;i<NumClusters;i++)
        {
		if (index==centroids[i]) return i;
                if (DistanceMatrix[index][centroids[i]]<min)
                {
                        min=DistanceMatrix[index][centroids[i]];
                        BestCluster=i;
                }
        }
        return BestCluster;
}

void AssignClusters(Matrix &DistanceMatrix, vector<int> &centroids, vector< vector<int> > &clusters)
{
        int BestCluster;
        int NumClusters=centroids.size();
        int Size=DistanceMatrix[0].size();
        clusters.clear();
        Safe2DAlloc(clusters, NumClusters, 0, "clusters in AssignClusters");

        for (int i=0;i<Size;i++)
        {
                BestCluster=FindBestCluster(centroids, DistanceMatrix, i);
		if (BestCluster<0 || BestCluster>NumClusters)
		{
			cout <<"ERROR: In AssignClusters BestCluster= "<<BestCluster<<endl;
			exit(EXIT_FAILURE);
		}
                SafePushBack(clusters[BestCluster], i, "clusters in AssignClusters");
        }
}

Real CalcSumDistances(int index, vector<int> &cluster, Matrix &DistanceMatrix)
{
        int Size=cluster.size();
        Real SumDistances=0;

        for (int i=0;i<Size;i++)
        {
                SumDistances+=DistanceMatrix[index][cluster[i]];
        }
        return SumDistances;
}

int FindCentroid(Matrix &DistanceMatrix, vector<int> &cluster)
{
        int Size=cluster.size();
        int centroid;
        Real MinSumDistances, SumDistances;

        MinSumDistances=CalcSumDistances(cluster[0], cluster, DistanceMatrix);
        centroid=cluster[0];
        for (int i=1;i<Size;i++)
        {
                SumDistances=CalcSumDistances(cluster[i], cluster, DistanceMatrix);
                if (SumDistances<MinSumDistances)
                {
                        MinSumDistances=SumDistances;
                        centroid=cluster[i];
                }
        }
        return centroid;
}

void FindCentroids(Matrix &DistanceMatrix, vector<int> &centroids, vector< vector<int> > &clusters)
{
        int ncentroid=centroids.size();
	int ncluster=clusters.size();
        int centroid;

	if (ncluster!=ncentroid)
	{
		string errorStr="ncluster= "+toStr(ncluster)+" ncentroid= "+toStr(ncentroid);
		error(errorStr, __LINE__, __FILE__);
	}
        for (int i=0;i<ncentroid;i++)
        {
                centroid=FindCentroid(DistanceMatrix, clusters[i]);
                centroids[i]=centroid;
        }
}

Real calcSumOfDistancesToCentroid(Matrix &DistanceMatrix, vector<int> &cluster, int centroid)
{
	int nincluster=cluster.size();
	Real sum=0;

	for (int i=0;i<nincluster;i++)
	{
		sum+=DistanceMatrix[cluster[i]][centroid];
	}
	return sum;
}

void calcCScore(Matrix &DistanceMatrix, vector<int> &cluster, int centroid, Real &cscore, Real &ave)
{
	int Size=DistanceMatrix.size();
	int nincluster=cluster.size();
	Real sum=0;

	sum=calcSumOfDistancesToCentroid(DistanceMatrix, cluster, centroid);
	ave=sum/Real(nincluster);
	cscore=Real(nincluster)/(ave*Real(Size));
}

void calcCScores(Matrix &DistanceMatrix, vector< vector<int> > &clusters, vector<int> &centroids, Vector &cScores, Vector &aves)
{
	int ncluster=clusters.size();

	SafeAlloc(cScores, ncluster, "cScores");
	SafeAlloc(aves, ncluster, "cScores");

	for (int i=0;i<ncluster;i++)
	{
		calcCScore(DistanceMatrix, clusters[i], centroids[i], cScores[i], aves[i]);
	}
}

Real calcClusteringScore(Matrix &DistanceMatrix, vector< vector<int> > &clusters, vector<int> &centroids)
{
	int nclusters=centroids.size();
	Real sum=0;

	for (int i=0;i<nclusters;i++)
	{
		sum+=calcSumOfDistancesToCentroid(DistanceMatrix, clusters[i], centroids[i]);
	}
	return sum;
}

void printClusters(vector<int> &centroids, vector< vector<int> > &clusters)
{
        int ncluster=centroids.size();
        int nincluster;
        for (int i=0;i<ncluster;i++)
        {
                nincluster=clusters[i].size();
                cout <<endl;
                cout <<"centroids["<<i<<"]= "<<centroids[i]<<endl;
                cout <<"nincluster= "<<nincluster<<endl;
                for (int j=0;j<nincluster;j++)
                {
                        cout <<clusters[i][j]<<"\t";
                        if ((j+1)%10==0) cout <<endl;
                }
                cout <<endl<<endl;
        }
}

void cluster(Matrix &DistanceMatrix, int NumClusters, vector<int> &centroids, vector< vector<int> > &clusters)
{
        int MaxIterations=1000;
        int Size=DistanceMatrix[0].size();
        AssignCentroidsAtRandom(Size, NumClusters, centroids);
        for (int i=0;i<MaxIterations;i++)
        {
                AssignClusters(DistanceMatrix, centroids, clusters);
                FindCentroids(DistanceMatrix, centroids, clusters);
        }
}

void clusterMultipleTimes(Matrix &DistanceMatrix, int NumClusters, vector<int> &centroids, vector< vector<int> > &clusters, int ntries)
{
	vector<int> bestCentroids;
	vector< vector<int> > bestClusters;
	Real score, bestScore;
	cluster(DistanceMatrix, NumClusters, bestCentroids, bestClusters);
	bestScore=calcClusteringScore(DistanceMatrix, bestClusters, bestCentroids);
	cout <<"bestScore= "<<bestScore<<endl;
	for (int i=0;i<ntries-1;i++)
	{
		cluster(DistanceMatrix, NumClusters, centroids, clusters);
		score=calcClusteringScore(DistanceMatrix, clusters, centroids);
		cout <<"Try number= "<<i<<" score= "<<score<<endl;
		if (score<bestScore)
		{
			bestScore=score;
			cout <<"bestScore= "<<bestScore<<endl;
			bestCentroids=centroids;
			bestClusters=clusters;
		}
	}
	centroids=bestCentroids;
	clusters=bestClusters;
}

bool areAllAssigned(vector<bool> &assigned)
{
	int Size=assigned.size();

	for (int i=0;i<Size;i++)
	{
		if (!assigned[i]) return false;
	}
	return true;
}

int countNumUnassigned(vector<bool> &assigned)
{
	int Size=assigned.size();
	int count=0;

	for (int i=0;i<Size;i++)
	{
		if (!assigned[i]) count++;
	}
	return count;
}

int findUnassigned(vector<bool> &assigned)
{
	int Size=assigned.size();

	for (int i=0;i<Size;i++)
	{
		if (!assigned[i]) return i;
	}
	return UNK_INT;
}

void setAssignedToFalse(vector<bool> &assigned)
{
	int Size=assigned.size();

	for (int i=0;i<Size;i++) assigned[i]=false;
}

void getAssigned(vector< vector<int> > &clusters, vector<bool> &assigned)
{
	int ncluster=clusters.size();
	int Size;

	setAssignedToFalse(assigned);
	for (int i=0;i<ncluster;i++)
	{
		Size=clusters[i].size();
		for (int j=0;j<Size;j++)
		{
			cout <<"i= "<<i<<" j= "<<j<<" clusters= "<<clusters[i][j]<<endl;
			assigned[clusters[i][j]]=true;
		}
	}
}

int findClosestPair(Matrix &distanceMatrix, vector<bool> &assigned)
{
	bool firstPair=true;
	int closest=UNK_INT;
	int Size=distanceMatrix.size();
	Real min=UNK_REAL;

	if (countNumUnassigned(assigned)==1) 
	{
		return findUnassigned(assigned);
	}
	for (int i=0;i<Size;i++)
	{
		if (!assigned[i])
		{
			for (int j=i+1;j<Size;j++)
			{
				if (!assigned[j])
				{
					if (firstPair)
					{
						min=distanceMatrix[i][j];
						closest=i;
						//cout <<"min= "<<min<<endl;
						firstPair=false;
					}
					if (distanceMatrix[i][j]<min)
					{
						min=distanceMatrix[i][j];
						//cout <<"min= "<<min<<endl;
						closest=i;
					}
				}
			}
		}
	}
	return closest;
}

int findNumWithinCutoff(Matrix &distanceMatrix, vector<bool> &assigned, Real cutoff, int index)
{
	int Size=distanceMatrix.size();
	int numWithinCutoff=0;

	for (int i=0;i<Size;i++)
	{
		if (i!=index && !assigned[i] && distanceMatrix[i][index]<cutoff)
		{
			numWithinCutoff++;
		}
	}
	return numWithinCutoff;
}

int findCenter(Matrix &distanceMatrix, vector<bool> &assigned, Real cutoff)
{
	int Size=distanceMatrix.size();
	int maxWithinCutoff=0;
	int numWithinCutoff;
	int center=UNK_INT;

	for (int i=0;i<Size;i++)
	{
		if (!assigned[i])
		{
			numWithinCutoff=findNumWithinCutoff(distanceMatrix, assigned, cutoff, i);
			if (numWithinCutoff>=maxWithinCutoff)
			{
				maxWithinCutoff=numWithinCutoff;
				center=i;
			}
		}
	}
	return center;
}

void assignToCluster(vector<int> &centroids, Matrix &distanceMatrix, vector<bool> &assigned, vector< vector<int> > &clusters, Real cutoff, int index)
{
	int ncentroids=centroids.size();
	int ncluster=clusters.size();
	int best;
	vector<int> temp;
	Real min;


	min=distanceMatrix[centroids[0]][index];
	best=0;
	for (int i=0;i<ncentroids;i++)
	{
		//cout <<endl;
		//cout <<"ncentroids= "<<ncentroids<<endl;
		//cout <<"ncluster= "<<ncluster<<endl;
		//cout <<"i= "<<i<<endl;
		//cout <<"index= "<<index<<endl;
		//cout <<"centroids= "<<centroids[i]<<endl;
		//cout <<"distanceMatrix= "<<distanceMatrix[centroids[i]][index]<<endl;
		if (distanceMatrix[centroids[i]][index]<min)
		{
			min=distanceMatrix[centroids[i]][index];
			best=i;
		}
		if (centroids[i]==index)
		{
			//cout <<"min= "<<min<<" best= "<<best<<endl;
		}
	}
	if (min<cutoff)
	{
		if (best<ncluster)
		{
			SafePushBack(clusters[best], index, "clusters");
		}
		else
		{
			SafeAlloc(temp, 1, "temp");
			temp[0]=index;
			SafePushBack(clusters, temp, "clusters");
		}
		assigned[index]=true;
	}
}

void assignToClusters(vector<int> &centroids, Matrix &distanceMatrix, vector<bool> &assigned, vector< vector<int> > &clusters, Real cutoff)
{
	int Size=distanceMatrix.size();
	//clusters.clear();
	//cout <<"centroids:"<<endl;
	//printVector(centroids);
	//cout <<endl;
	for (int i=0;i<Size;i++)
	{
		if (!assigned[i])
		{
			assignToCluster(centroids, distanceMatrix, assigned, clusters, cutoff, i);
		}
	}
}

void findClosestAndAssign(vector<int> &centroids, Matrix &distanceMatrix, vector<bool> &assigned, vector< vector<int> > &clusters, Real cutoff)
{
	int closest;

	while (true)
	{
		if (areAllAssigned(assigned)) break;
		closest=findClosestPair(distanceMatrix, assigned);
		SafePushBack(centroids, closest, "centroids");
		assignToClusters(centroids, distanceMatrix, assigned, clusters, cutoff);
	}
}

void findCenterAndAssign(vector<int> &centroids, Matrix &distanceMatrix, vector<bool> &assigned, vector< vector<int> > &clusters, Real cutoff)
{
	int closest;

	while (true)
	{
		if (areAllAssigned(assigned)) break;
		closest=findCenter(distanceMatrix, assigned, cutoff);
		SafePushBack(centroids, closest, "centroids");
		assignToClusters(centroids, distanceMatrix, assigned, clusters, cutoff);
	}
}

void clusterCutoff(Matrix &distanceMatrix, Real cutoff, vector<int> &centroids, vector< vector<int> > &clusters)
{
	vector<bool> assigned;
	int Size=distanceMatrix.size();
	int maxIterations=10;

	SafeAlloc(assigned, Size, "assigned");
	//findClosestAndAssign(centroids, distanceMatrix, assigned, clusters, cutoff);
	findCenterAndAssign(centroids, distanceMatrix, assigned, clusters, cutoff);
	for (int i=0;i<maxIterations;i++)
	{
		FindCentroids(distanceMatrix, centroids, clusters);
		getAssigned(clusters, assigned);
		assignToClusters(centroids, distanceMatrix, assigned, clusters, cutoff);
		if (!areAllAssigned(assigned))
		{
			//findClosestAndAssign(centroids, distanceMatrix, assigned, clusters, cutoff);
			findCenterAndAssign(centroids, distanceMatrix, assigned, clusters, cutoff);
		}
	}
	//printClusters(centroids, clusters);
	//endProgram(__LINE__, __FILE__);
}

#endif
