# include <vector>
# include <iostream>

using namespace std;

# include "../library/MathUtils.h"
# include "../library/normalize.h"
# include "../library/MinMax.h"
# include "../library/time.h"
# include "../library/VectorManip.h"
# include "XRayStruct.h"
# include "XRay.h"

struct MltfStruct
{
	vector<bool> centric;
	Vector fObs, fCalc, iStdDev, fError, otherError;
};

struct CalcObsStruct
{
	XRayStruct xray, expXRay;
};

Real calcWoolfson(Real fObs, Real fCalc, Real fVar)
{
	//Woolfson, M. M. (1956). Acta Cryst. 9, 804-810.
	Real w;

	w=sqrt(2.0/(pi*fVar))*cosh(fObs*fCalc/fVar)/exp((fObs*fObs+fCalc*fCalc)/(2.0*fVar));
	return w;
}

Real calcLogWoolfson(Real fObs, Real fCalc, Real fVar)
{
	//Woolfson, M. M. (1956). Acta Cryst. 9, 804-810.
	Real w, x, cutoff=15.0;

	x=fObs*fCalc/fVar;
	w=0.5*log(2.0/(pi*fVar));
	if (x<cutoff) w+=log(cosh(x));
	else w+=x-log(2);
	w-=(fObs*fObs+fCalc*fCalc)/(2.0*fVar);
	return w;
}

Real calcRice(Real fObs, Real fCalc, Real fVar)
{
	//Sim, G. A. (1959). Acta Cryst. 12, 813-815
	//The distributio of phase angles for structures...
	Real r, i0;

	i0=calcBesselI0(2.0*fObs*fCalc/fVar);
	r=(2.0*fObs/fVar)*i0/exp((fObs*fObs+fCalc*fCalc)/fVar);
	return r;
}

Real calcLogRice(Real fObs, Real fCalc, Real fVar)
{
	//Sim, G. A. (1959). Acta Cryst. 12, 813-815
	//The distributio of phase angles for structures...
	Real r, x;

	if (fObs==0) fObs=1.0e-8;
	x=2.0*fObs*fCalc/fVar;
	r=log(2.0);
	r+=log(fObs);
	r-=log(fVar);
	r+=calcLogBesselI0(x);
	r-=(fObs*fObs+fCalc*fCalc)/fVar;

	return r;
}

Real calcLogRice(Real realAve, Real imagAve, Real realStdDev, Real imagStdDev, Real fObs)
{
	Real fAve, fVar;

	fAve=sqrt(realAve*realAve+imagAve*imagAve);
	fVar=2.0*realStdDev*imagStdDev;

	return calcLogRice(fObs, fAve, fVar);
}

Real calcMLRF0_NoScale(vector<Real> &fObs, vector<Real> &fCalc, vector<Real> &fError, vector<bool> &centric)
{
	//J. Appl. Cryst. (2007). 40, 658-674
	int nobs=fObs.size();
	int ncalc=fCalc.size();
	int nerror=fError.size();
	int ncentric=centric.size();
	Real mlrf0=0, fVar;
	Real sumFObs2, sumFCalc2;

	if (nobs==0 || nobs!=ncalc || nobs!=nerror || nobs!=ncentric)
	{
		string errorStr="nobs= "+IntToStr(nobs)+" ncalc= ";
		errorStr+=IntToStr(ncalc)+" nerror= "+IntToStr(nerror);
		errorStr+=" ncentric= "+IntToStr(ncentric);
		error(errorStr, __LINE__, __FILE__);
	}
	
	sumFObs2=calcSumSqr(fObs);
	cout <<"sumFObs2= "<<sumFObs2<<endl;
	sumFCalc2=calcSumSqr(fCalc);
	cout <<"sumFCalc2= "<<sumFCalc2<<endl;

	for (int i=0;i<nobs;i++)
	{
		cout <<"i= "<<i<<endl;
		fVar=sumFObs2+fCalc[i]*fCalc[i]-sumFCalc2+fError[i];
		if (centric[i]) mlrf0+=log(calcWoolfson(fObs[i], 0, fVar));
		else mlrf0+=log(calcRice(fObs[i], 0, fVar));
	}
	return mlrf0;
}

Real calcMLRF0(vector<Real> &fObs, vector<Real> &fCalc, vector<Real> &fError, vector<bool> &centric)
{
	//Get scale that minimizes difference of squares.  Change later?
	Real scale, mlrf0;
	Vector fCalcScaled;
	cout <<"Before calcScaleForSqrDiff"<<endl;
	scale=calcScaleForSqrDiff(fCalc, fObs);
	cout <<"scale= "<<scale<<endl;
	multiply(scale, fCalc, fCalcScaled);
	cout <<"Before calcMLRF0_NoScale"<<endl;
	mlrf0=calcMLRF0_NoScale(fObs, fCalcScaled, fError, centric);
	cout <<"mlrf0"<<endl;
	return mlrf0;
}

Real calcMLRF0(XRayStruct &xray, XRayStruct &expXRay)
{
	Real mlrf0;
	calcAmplitude(expXRay);
	calcAmplitude(xray);
	mlrf0=calcMLRF0(xray.F, expXRay.F, expXRay.Ferror, xray.centric);
	cout <<"mlrf0= "<<mlrf0<<endl;
	return mlrf0;
}

Vector calcLuzzatiD(Real rmsd, Vector &qMag)
{
	int npoint=qMag.size();
	Real sigma;
	Vector D;

	SafeAlloc(D, npoint, "D");
	for (int i=0;i<npoint;i++)
	{
		sigma=2.0*rmsd*rmsd/3.0;
		D[i]=1.0/exp(pi*pi*sigma*qMag[i]*qMag[i]);	
	}
	return D;
}

Real calcLuzzatiDNumeric(Real rmsd, Real sx, Real sy, Real sz)
{
	int niter=100000;
	Real sum=0;
	//Real sigma=rmsd*sqrt(2.0/3.0);
	Real sigma=rmsd*sqrt(1.0/3.0);
	Real x, y, z;

	for (int i=0;i<niter;i++)
	{
		x=randGaussian(sigma);
		y=randGaussian(sigma);
		z=randGaussian(sigma);
		sum+=cos(2.0*pi*(x*sx+y*sy+z*sz));
	}
	return sum/Real(niter);
}

Vector calcLuzzatiDNumeric(Real rmsd, vector<PosStruct> &q)
{
	int npoint=q.size();
	Vector D;

	SafeAlloc(D, npoint, "D");
	for (int i=0;i<npoint;i++)
	{
		D[i]=calcLuzzatiDNumeric(rmsd, q[i].Pos[X], q[i].Pos[Y], q[i].Pos[Z]);
	}
	return D;
}


void compareLuzzatiD(Real rmsd, XRayStruct &xray)
{
	Vector D, DNumeric;
	int Size;

	D=calcLuzzatiD(rmsd, xray.qMag);
	DNumeric=calcLuzzatiDNumeric(rmsd, xray.q);
	
	Size=D.size();

	for (int i=0;i<Size;i++)
	{
		cout <<"i= "<<i<<" D= "<<D[i]<<" DNumeric= "<<DNumeric[i]<<" qMag= "<<xray.qMag[i]<<endl;
	}
	endProgram(__LINE__, __FILE__);
}

void plotLogWoolfson()
{
	Real fObs=1.0;
	Real D=1.0;
	Real fVar=10.0;
	Real inc=0.01;
	Real fMax=4.0;
	Real mltf;

	for (Real fCalc=0;fCalc<fMax;fCalc+=inc)
	{
		mltf=calcLogWoolfson(fObs, D*fCalc, fVar);
		cout <<"fCalc= "<<fCalc<<" mltf= "<<mltf<<endl;
	}
	endProgram(__LINE__, __FILE__);
}

void plotLogRice()
{
	Real fObs=1.0;
	//Real fError=0.1;
	Real D=1.0;
	Real fVar=10.0;
	Real inc=0.01;
	Real fMax=4.0;
	Real mltf;

	for (Real fCalc=0;fCalc<fMax;fCalc+=inc)
	{
		mltf=calcLogRice(fObs, D*fCalc, fVar);
		cout <<"fCalc= "<<fCalc<<" mltf= "<<mltf<<" mltf2= "<<log(calcRice(fObs, D*fCalc, fVar))<<endl;
	}
	endProgram(__LINE__, __FILE__);
}

vector<ProteinStruct> addGaussianNoiseToProteins(vector<ProteinStruct> &Proteins, Real rmsd)
{
	int natom=Proteins[0].Atoms.size();
	int nprot=Proteins.size();
	Real sigma;
	ProteinStruct tempProtein;
	vector<ProteinStruct> tempProteins;

	sigma=rmsd/sqrt(3.0);
	tempProtein=Proteins[0];
	for (int i=0;i<natom;i++)
	{
		tempProtein.Atoms[i].x+=randGaussian(sigma);
		tempProtein.Atoms[i].y+=randGaussian(sigma);
		tempProtein.Atoms[i].z+=randGaussian(sigma);
	}
	SafeAlloc(tempProteins, tempProtein, nprot, "tempProteins");
	return tempProteins;
}

void addGaussianNoiseToProteins(vector<ProteinStruct> &Proteins, Real rmsd, vector<ProteinStruct> &tempProteins)
{
	int natom=Proteins[0].Atoms.size();
	int nprot2=tempProteins.size();
	Real sigma;

	sigma=rmsd/sqrt(3.0);

	if (nprot2==0)
	{
		SafePushBack(tempProteins, Proteins[0], "tempProteins");
	}
	for (int i=0;i<natom;i++)
	{
		tempProteins[0].Atoms[i].x=Proteins[0].Atoms[i].x+randGaussian(sigma);
		tempProteins[0].Atoms[i].y=Proteins[0].Atoms[i].y+randGaussian(sigma);
		tempProteins[0].Atoms[i].z=Proteins[0].Atoms[i].z+randGaussian(sigma);
	}
}

void keepTrackOfError(XRayStruct &xray, Vector &ave, Vector &aveOfSqrs)
{
	int npoint=xray.i.size();

	for (int i=0;i<npoint;i++)
	{
		//ave[i]+=xray.i[i];
		//aveOfSqrs[i]+=xray.i[i]*xray.i[i];
		ave[i]+=sqrt(xray.i[i]);
		aveOfSqrs[i]+=xray.i[i];
	}
}

void calcRealImagDistributionDueToRmsd(vector<ProteinStruct> &Proteins, Real rmsd, XRayStruct xray, XRayStruct &expXRay, Matrix &reals, Matrix &imags)
{
	int niter=10000;
	Vector real, imag;
	vector<ProteinStruct> tempProteins;
	
	real=xray.real;
	imag=xray.imag;
	for (int i=0;i<niter;i++)
	{
		tempProteins=addGaussianNoiseToProteins(Proteins, rmsd);
		calcSymmetryScattering(xray, tempProteins);
		SafePushBack(reals, subtractVectors(xray.real, real), "reals");
		SafePushBack(imags, subtractVectors(xray.imag, imag), "imags");
	}	
}

void calcRealImagHistogramsDueToRmsd(vector<ProteinStruct> &Proteins, Real rmsd, XRayStruct xray, XRayStruct &expXRay, Matrix &reals, Matrix &imags, Vector &histogram, Real binSize, Real minValue, int nbin)
{
	Vector realHistogram, imagHistogram;
	Matrix realst, imagst;

	cout <<"In calcRealImagHistogramsDueToRmsd"<<endl;
	calcRealImagDistributionDueToRmsd(Proteins, rmsd, xray, expXRay, reals, imags);
	
	transpose(reals, realst);
	transpose(imags, imagst);
	calcHistogram(realst[0], realHistogram, binSize, minValue, nbin);
	calcHistogram(imagst[0], imagHistogram, binSize, minValue, nbin);
	printHistogram(realHistogram, binSize, minValue, "Real", "Counts");
	printHistogram(imagHistogram, binSize, minValue, "Imag", "Counts");
}

Matrix calcAmplitudeDistributionDueToRmsd(vector<ProteinStruct> &Proteins, Real rmsd, XRayStruct xray, XRayStruct &expXRay)
{
	int niter=1000;
	int npoint=xray.miller.size();
	Matrix fs, reals, imags;
	Matrix fsTranspose;
	Matrix realTranspose, imagTranspose;
	vector<ProteinStruct> tempProteins;
	
	Safe2DAlloc(fs, niter, npoint, "fs");
	Safe2DAlloc(reals, niter, npoint, "reals");
	Safe2DAlloc(imags, niter, npoint, "imags");

	cout <<"In calcAmplitudeDistributionDueToRmsd"<<endl;

	for (int i=0;i<niter;i++)
	{
		addGaussianNoiseToProteins(Proteins, rmsd, tempProteins);
		calcSymmetryScattering(xray, tempProteins);
		//correctIntensity(xray);
		//fixBFactor(xray, expXRay);
		calcAmplitude(xray);
		fs[i]=xray.F;
		reals[i]=xray.real;
		imags[i]=xray.imag;
	}	

	transpose(fs, fsTranspose);
	transpose(reals, realTranspose);
	transpose(imags, imagTranspose);

	Real r, fAve;
	Real realAve, imagAve;
	Real realStdDev, imagStdDev;

	for (int i=0;i<npoint;i++)
	{
		r=calcPearson(realTranspose[i], imagTranspose[i]);
		realAve=calcAverage(realTranspose[i]);
		imagAve=calcAverage(imagTranspose[i]);
		realStdDev=calcStandardDeviationSample(realTranspose[i]);
		imagStdDev=calcStandardDeviationSample(imagTranspose[i]);
		fAve=calcAverage(fsTranspose[i]);
		cout <<"i= "<<i<<" realAve= "<<realAve<<" imagAve= "<<imagAve<<" realStdDev= "<<realStdDev<<" imagStdDev= "<<imagStdDev<<" fAve= "<<fAve<<" r= "<<r<<" centric= "<<xray.centric[i]<<endl;
	}

	return fs;
}

Vector calcErrorDueToRmsd(vector<ProteinStruct> &Proteins, Real rmsd, XRayStruct xray, XRayStruct &expXRay)
{
	int niter=100;
	int npoint=xray.i.size();
	Vector errors, ave, aveOfSqrs, originalI;
	vector<ProteinStruct> tempProteins;

	calcSymmetryScattering(xray, Proteins);
	correctIntensity(xray);
	fixBFactor(xray, expXRay);
	originalI=xray.i;
	npoint=xray.i.size();
	SafeAlloc(errors, npoint, "errors");
	SafeAlloc(ave, npoint, "ave");
	SafeAlloc(aveOfSqrs, npoint, "aveOfSqrs");


	for (int i=0;i<niter;i++)
	{
		tempProteins=addGaussianNoiseToProteins(Proteins, rmsd);
		calcSymmetryScattering(xray, tempProteins);
		correctIntensity(xray);
		fixBFactor(xray, expXRay);
		keepTrackOfError(xray, ave, aveOfSqrs);
	}	

	for (int i=0;i<npoint;i++)
	{
		errors[i]=sqrt(aveOfSqrs[i]/Real(niter)-ave[i]*ave[i]/(Real(niter)*Real(niter)));		
		cout <<"i= "<<i<<" intensity= "<<xray.i[i]<<" errors= "<<errors[i]<<" qMag= "<<xray.qMag[i]<<" centric= "<<xray.centric[i]<<" originalI= "<<originalI[i]<<" ave= "<<ave[i]/Real(niter)<<" h= "<<xray.miller[i].Pos[X]<<" k= "<<xray.miller[i].Pos[Y]<<" l= "<<xray.miller[i].Pos[Z]<<endl;
	}
	return errors;
}

Real calcMLTF_NoScale(Vector &fObs, Vector &fCalc, Vector &iStdDev, Vector &fError, Vector &otherError, vector<bool> &centric)
{
	//J. Appl. Cryst. (2007). 40, 658-674
	int nobs=fObs.size();
	int ncalc=fCalc.size();
	int nerror=fError.size();
	int ncentric=centric.size();
	Real mltf=0, fVar;
	timeval start, end;

	//plotLogRice();
	plotLogWoolfson();
	if (nobs==0 || nobs!=ncalc || nobs!=nerror || nobs!=ncentric)
	{
		string errorStr="nobs= "+IntToStr(nobs)+" ncalc= ";
		errorStr+=" nerror= "+IntToStr(nerror);
		errorStr+=" ncentric= "+IntToStr(ncentric);
		error(errorStr, __LINE__, __FILE__);
	}
	gettimeofday(&start, NULL);
	for (int i=0;i<nobs;i++)
	{
		//fVar=iStdDev[i]+fError[i]*fError[i];
		fVar=iStdDev[i]*iStdDev[i]+fError[i]*fError[i]+otherError[i]*otherError[i];
		if (fVar==0 || isnan(fVar)) fVar=fObs[i]*0.1;
		//if (centric[i]) mltf+=calcLogWoolfson(fObs[i], fCalc[i], fVar);
		//else mltf+=calcLogRice(fObs[i], fCalc[i], fVar);	
		cout <<"fObs= "<<fObs[i]<<" fVar= "<<fVar<<endl;
		if (centric[i]) mltf+=calcLogWoolfson(fObs[i], fObs[i], fVar);
		else mltf+=calcLogRice(fObs[i], fObs[i], fVar);	
		if (centric[i]) cout <<"mltfPoint= "<<calcLogWoolfson(fObs[i], fObs[i], fVar)<<endl;
		else cout <<"mltfPoint= "<<calcLogRice(fObs[i], fObs[i], fVar)<<endl;	
	}
	gettimeofday(&end, NULL);
	endProgram(__LINE__, __FILE__);
	mltf=-mltf;
	return mltf;
}

Real calcOtherErrorScale(Vector &fObs, Vector &fCalc, Vector &iStdDev, Vector &fError, Vector &otherError, Real scale)
{
	int npoint=fObs.size();
	Real total=0;

	for (int i=0;i<npoint;i++)
	{
		total+=(abs(scale*fCalc[i]-fObs[i])-fError[i]-scale*iStdDev[i])/otherError[i];
	}
	return total/Real(npoint);
}

Real calcMLTF(Vector &fObs, Vector fCalc, Vector iStdDev, Vector &fError, Vector &otherError, vector<bool> &centric, Real scale, Real scaleOtherError)
{
	int npoint=fCalc.size();

	for (int i=0;i<npoint;i++)
	{
		fCalc[i]*=scale;
		//iStdDev[i]*=scale*scale;
		iStdDev[i]*=scale;
		otherError[i]*=scaleOtherError;
	}
	return calcMLTF_NoScale(fObs, fCalc, iStdDev, fError, otherError, centric);
}

Real calcMLTF(Vector scale, MltfStruct args)
{
	return calcMLTF(args.fObs, args.fCalc, args.iStdDev, args.fError, args.otherError, args.centric, scale[0], 1.0);
}

Real calcMLTFBothScales(Vector scale, MltfStruct args)
{
	return calcMLTF(args.fObs, args.fCalc, args.iStdDev, args.fError, args.otherError, args.centric, scale[0], scale[1]);
}

Real calcMLTFOptimizeScale(XRayStruct &xray, XRayStruct &expXRay)
{
	Real dScoreCriteria=0.1;
	Real maxHours=1.0/600.0;
	Real Scale, mltf;
	Vector scale;
	MltfStruct args;
	Real (*func_to_minimize)(Vector, MltfStruct)=calcMLTF;

	args.fObs=expXRay.F;
	args.fCalc=xray.F;
	args.iStdDev=xray.iStdDev;
	args.fError=expXRay.Ferror;
	args.otherError=xray.otherError;
	args.centric=xray.centric;
	
	Scale=calcScaleForSqrDiff(xray.F, expXRay.F);
	SafePushBack(scale, Scale, "scale");

	mltf=optimizeCoefficients(scale, args, func_to_minimize, dScoreCriteria, maxHours);

	return mltf;
}

Real calcMLTFOptimizeBothScales(XRayStruct &xray, XRayStruct &expXRay)
{
	Real dScoreCriteria=0.000000001;
	Real maxHours=1.0/2.0;
	Real Scale, mltf;
	Real otherScale;
	Vector scale;
	MltfStruct args;
	Real (*func_to_minimize)(Vector, MltfStruct)=calcMLTFBothScales;

	args.fObs=expXRay.F;
	args.fCalc=xray.F;
	args.iStdDev=xray.iStdDev;
	args.fError=expXRay.Ferror;
	args.otherError=xray.otherError;
	args.centric=xray.centric;
	
	Scale=calcScaleForSqrDiff(xray.F, expXRay.F);
	otherScale=calcOtherErrorScale(expXRay.F, xray.F, xray.iStdDev, expXRay.Ferror, xray.otherError, Scale);
	SafePushBack(scale, Scale, "scale");
	SafePushBack(scale, otherScale, "scale");

	
	mltf=optimizeCoefficients(scale, args, func_to_minimize, dScoreCriteria, maxHours);
	//cout <<"Before conjugateGradient"<<endl;
	//mltf=conjugateGradient(scale, args, func_to_minimize, dScoreCriteria, maxHours);
	cout <<"mltf= "<<mltf<<endl;
	endProgram(__LINE__, __FILE__);
	return mltf;
}

Real calcMltfZScore(vector<ProteinStruct> &Proteins, XRayStruct &xray, XRayStruct &expXRay, Matrix &fs, Real rmsd)
{
	int nplacement=fs.size();
	Real mltf;
	Vector mltfs, zscores;

	cout <<"In calcMlftZScore"<<endl;
	xray.iStdDev=calcErrorDueToRmsd(Proteins, rmsd, xray, expXRay);
	
	for (int i=0;i<nplacement;i++)
	{
		xray.F=fs[i];
		cout <<"Before calcMLTFOptimizeBothScales"<<endl;
		mltf=calcMLTFOptimizeBothScales(xray, expXRay);
		cout <<"After calcMLTFOptimizeBothScales"<<endl;
		SafePushBack(mltfs, mltf, "mltfs");
	}
	calcZScore(mltfs, zscores);
	return findMin(zscores);
}

Real calcMLTFOptimizeRmsd(vector<ProteinStruct> &Proteins, XRayStruct &xray, XRayStruct &expXRay, Matrix &fs, Real rmsd)
{
	Real zScore, oldZScore;
	Real step=0.1;
	Real criteria=0.01;
	Real bestScore, bestRmsd;

	oldZScore=calcMltfZScore(Proteins, xray, expXRay, fs, rmsd);	
	bestScore=oldZScore;
	bestRmsd=rmsd;

	cout <<"rmsd= "<<rmsd<<" zScore= "<<oldZScore<<endl;
	while (true)
	{
		if (rmsd<0) return 0;
		rmsd=bestRmsd+step;
		zScore=calcMltfZScore(Proteins, xray, expXRay, fs, rmsd);
		cout <<"rmsd= "<<rmsd<<" zScore= "<<zScore<<endl;
		if (bestScore>zScore) 
		{
			step*=1.1;
			bestRmsd=rmsd;
			bestScore=zScore;
		}
		else 
		{
			step*=-0.5;
		}
		oldZScore=zScore;
		if (abs(step)<criteria) break;		
	}
	cout <<"rmsd optimized for maximum likelihood: "<<bestRmsd<<endl;
	return bestRmsd;
}

Real calcMLTF(Vector &fObs, Vector fCalc, Vector iStdDev, Vector &fError, Vector &otherError, vector<bool> &centric)
{
	Real scale=calcScaleForSqrDiff(fCalc, fObs);

	return calcMLTF(fObs, fCalc, iStdDev, fError, otherError, centric, scale, 1.0);
}

Real calcMLTF(XRayStruct &xray, XRayStruct &expXRay, Real rmsd)
{
	Real mltf;
	//plotLogWoolfson();
	//compareLuzzatiD(rmsd, xray);
	//xray.D=calcLuzzatiD(rmsd, xray.qMag);
	mltf=calcMLTF(expXRay.F, xray.F, xray.iStdDev, expXRay.Ferror, xray.otherError, expXRay.centric);
	//cout <<"mltf= "<<mltf<<endl;
	//endProgram(__LINE__, __FILE__);
	return mltf;
}

void calcMeanIntensity(XRayStruct &xray, Vector &meanI)
{
	//Given the amplitudes of symmetry mates calculates the
	//average intensity over a random walk of phases.
	//Does so by evaluating the following integral.
	//inegral[ (sigma(r_i*cos(theta_i))^2 + (sigma(r_i*sin(theta_i))^2 ] dtheta
	int nprot, nsym, npoint;

	Get3DVectorSize(xray.complexSymReal, nprot, nsym, npoint, "complexSym");


	SafeAlloc(meanI, npoint, "rootMeanI");
	for (int i=0;i<npoint;i++)
	{
		meanI[i]=0;
		for (int j=0;j<nsym;j++)
		{
			for (int k=0;k<nprot;k++)
			{
				meanI[i]+=xray.complexSymReal[k][j][i]*xray.complexSymReal[k][j][i];
				meanI[i]+=xray.complexSymImag[k][j][i]*xray.complexSymImag[k][j][i];
			}
		}
	}
}

void calcRootMeanIntensity(XRayStruct &xray, Vector &rootMeanI)
{
	//Given the amplitudes of symmetry mates calculates the
	//average intensity over a random walk of phases.
	//Does so by evaluating the following integral.
	//inegral[ (sigma(r_i*cos(theta_i))^2 + (sigma(r_i*sin(theta_i))^2 ] dtheta
	int npoint;

	calcMeanIntensity(xray, rootMeanI);
	npoint=rootMeanI.size();
	for (int i=0;i<npoint;i++)
	{
		rootMeanI[i]=sqrt(rootMeanI[i]);
	}
}

void calcAverageSquaredIntensity(XRayStruct &xray, Vector &aveSqrIntensity)
{
	//Calculates the average squared intensity over a random walk of phases.
	//Does so by evaluating the following integral.
	//inegral[ (sigma(r_i*cos(theta_i))^2 + (sigma(r_i*sin(theta_i))^2 ]^2 dtheta
	int nsym=xray.symReal.size();
	int npoint=0;
	Real sum;
	Real a2, a4, b2;
	//Real a2, a4, b2, b4;

	if (nsym>0)
	{
		npoint=xray.symReal[0].size();
	}
	else
	{
		string errorStr="No info on amplitudes of symmetry mates.";
		error(errorStr, __LINE__, __FILE__);
	}

	SafeAlloc(aveSqrIntensity, npoint, "aveSqrIntensity");
	for (int i=0;i<npoint;i++)
	{
		sum=0;
		for (int j=0;j<nsym;j++)
		{
			a2=xray.symReal[j][i]*xray.symReal[j][i];
			a2+=xray.symImag[j][i]*xray.symImag[j][i];
			a4=a2*a2;
			//sum+=3.0*a4/8.0;
			//sum+=a4/8.0;
			sum+=a4*0.5;
			for (int k=j;k<nsym;k++)
			{
				b2=xray.symReal[j][i]*xray.symReal[j][i];
				b2+=xray.symImag[j][i]*xray.symImag[j][i];
				//b4=a2*a2;
				//sum+=3.0*(real2a*real2b+imag2a*imag2b)/2.0;
				//sum+=(real2a*real2b+imag2a*imag2b)/2.0;
				sum+=2.5*a2*b2;
			}
		}
		aveSqrIntensity[i]=sum;
	}
}

void calcAverageAndStdDevIntensity(XRayStruct &xray, Vector &rootMeanIntensity, Vector &stdDevIntensity)
{
	int npoint;
	int index=22;
	Real variance;
	Vector rootMeanI, aveSqrIntensity;

	cout <<"Before calcRootMeanIntensity(xray, rootMeanI)"<<endl;
	calcRootMeanIntensity(xray, rootMeanI);
	cout <<"Before calcAverageSquaredIntensity(xray, aveSqrIntensity)"<<endl;
	calcAverageSquaredIntensity(xray, aveSqrIntensity);	
	cout <<"After calcAverageSquaredIntensity(xray, aveSqrIntensity)"<<endl;
	npoint=rootMeanI.size();
	SafeAlloc(stdDevIntensity, npoint);
	for (int i=0;i<npoint;i++)
	{
		variance=aveSqrIntensity[i]-rootMeanI[i]*rootMeanI[i];
		stdDevIntensity[i]=sqrt(variance);
	}
	cout <<"rootMeanI["<<index<<"]= "<<rootMeanI[index]<<" stdDevIntensity= "<<stdDevIntensity[index]<<endl;
}

Real calcRandomIntensity(XRayStruct &xray, int index)
{
	int nsym, nprot, npoint;
	Real real=0, imag=0, dp;

	Get3DVectorSize(xray.complexSymReal, nprot, nsym, npoint, "complexSymReal");
	for (int i=0;i<nprot;i++)
	{
		for (int j=0;j<nsym;j++)
		{
			dp=randDouble(0, 2.0*pi);
			real+=cos(dp)*xray.complexSymReal[i][j][index];
			real-=sin(dp)*xray.complexSymImag[i][j][index];
			imag+=sin(dp)*xray.complexSymReal[i][j][index];
			imag+=cos(dp)*xray.complexSymImag[i][j][index];
		}
	}
	//endProgram(__LINE__, __FILE__);
	//return sqrt(real*real+imag*imag);
	return real*real+imag*imag;
}

void calcIntensityHistogram(XRayStruct &xray, Vector &histogram)
{
	int max=20000000;
	int index=22;
	int Size;
	Real intensity, binSize=100.0, x=0;
	Vector intensities;

	for (int i=0;i<max;i++)
	{
		intensity=calcRandomIntensity(xray, index);
		SafePushBack(intensities, intensity, "intensities");
	}
	calcHistogram(intensities, histogram, binSize);
	Size=histogram.size();
	int nmax=10000;
	int end=min(nmax, Size);
	for (int i=0;i<end;i++)
	{
		cout <<x<<"\t"<<histogram[i]<<endl;
		x+=binSize;
	}
	for (int i=Size;i<nmax;i++)
	{
		cout <<x<<"\t"<<0<<endl;
		x+=binSize;

	}
}

void calcIntensityHistogram(ProteinStruct &Protein, XRayStruct &xray, Vector &histogram)
{
	calcSymmetryScattering(xray, Protein, 0);
	calcIntensityHistogram(xray, histogram);
}

void calcMinMaxIntensity(XRayStruct &xray, int index, Real &min, Real &max, Vector &amplitudes)
{
	int nprot=xray.complexSymReal.size();
	int nsym=xray.complexSymReal[0].size();
	int nmax, count=0;
	Real amplitude, largest;

	for (int i=0;i<nsym;i++)
	{
		for (int j=0;j<nprot;j++)
		{
			amplitude=xray.complexSymReal[j][i][index]*xray.complexSymReal[j][i][index];
			amplitude+=xray.complexSymImag[j][i][index]*xray.complexSymImag[j][i][index];
			amplitude=sqrt(amplitude);
			amplitudes[count]=amplitude;
			count++;
		}
	}
	nmax=findMax(amplitudes, largest);
	max=largest;
	min=largest;
	for (int i=0;i<nsym;i++)
	{
		if (i!=nmax)
		{
			max+=amplitudes[i];
			min-=amplitudes[i];
		}
	}
	if (min<0) min=0;
}

Real calcRotationPossibleScore(XRayStruct &xray, XRayStruct &expXRay)
{
	int npoint;
	//int npoint1, npoint2;
	int nprot=xray.complexSymReal.size();
	int nsym;
	Real score=0, min, max, scale;
	Vector amplitudes, meanI;

	//calcAmplitude(xray);
	//calcAmplitude(expXRay);
	npoint=xray.miller.size();
	//npoint1=xray.F.size();	
	//npoint2=expXRay.F.size();	

	if (nprot==0)
	{
		error("nprot= 0", __LINE__, __FILE__);
	}
/*
	if (npoint1!=npoint2)
	{
		string errorStr="npoint1= "+toStr(npoint1)+" npoint2= ";
		errorStr+=toStr(npoint2);
		error(errorStr, __LINE__, __FILE__);
	}
*/
	//calcRootMeanIntensity(xray, rootMeanI);
	calcMeanIntensity(xray, meanI);
	scale=calcScaleForSqrDiff(meanI, expXRay.i);
	scale=sqrt(scale);
	//scale=calcScaleForSqrDiff(xray.F, expXRay.F);
	nsym=xray.complexSymReal[0].size();
	SafeAlloc(amplitudes, nprot*nsym, "amplitudes");
	for (int i=0;i<npoint;i++)
	{
		calcMinMaxIntensity(xray, i, min, max, amplitudes);
		min*=scale;
		max*=scale;
		//cout <<min<<"\t"<<max<<"\t"<<expXRay.i[i]<<endl;
		if (expXRay.i[i]>max*max)
		{
			score+=(max*max-expXRay.i[i])*(max*max-expXRay.i[i]);
		}
		if (expXRay.i[i]<min*min)
		{
			score+=(min*min-expXRay.i[i])*(min*max-expXRay.i[i]);
		}
/*
		if (expXRay.i[i]>max)
		{
			score+=(max-expXRay.F[i])*(max-expXRay.F[i]);
		}
		if (expXRay.i[i]<min)
		{
			score+=(min-expXRay.F[i])*(min-expXRay.F[i]);
		}
*/
	}
	return score;
}

Real calcRotationPossibleScore(ProteinStruct &Protein, XRayStruct &xray, XRayStruct &expXRay)
{
	removeNonoverlapping(xray, expXRay);	
	calcSymmetryScattering(xray, Protein, 0);
	return calcRotationPossibleScore(xray, expXRay);

}

Real calcAnalyticalLikelihood(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, Real f0)
{
        Real stdDev, mean;
        Real normalization;
        Real sum, product;
	Real prob=0;

        realStdDev=2.0*realStdDev*realStdDev;
        imagStdDev=2.0*imagStdDev*imagStdDev;
        stdDev=sqrt(realStdDev*imagStdDev);
        mean=sqrt(realMean*realMean+imagMean*imagMean);
        normalization=sqrt(pi)*exp(mean*mean/stdDev)*erf(mean/sqrt(stdDev))*sqrt(stdDev);
        sum=f0*f0+realMean*realMean+imagMean*imagMean;
        sum/=stdDev;
        product=2.0*f0*sqrt(realMean*realMean+imagMean*imagMean)/stdDev;
        if (abs(product-sum)<50.0) prob+=exp(product-sum);
        if (abs(product+sum)<50.0) prob-=exp(-product-sum);
        prob/=normalization;
        prob*=exp((realMean*realMean+imagMean*imagMean)/stdDev);

	return prob;
}

Real calcAnalyticalLogLikelihood(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, Real f0)
{
        Real stdDev, mean;
        Real normalization;
        Real sum, product;
	Real prob=0;

        realStdDev=2.0*realStdDev*realStdDev;
        imagStdDev=2.0*imagStdDev*imagStdDev;
        stdDev=sqrt(realStdDev*imagStdDev);
        mean=sqrt(realMean*realMean+imagMean*imagMean);
	normalization=0.5*log(pi);
	normalization+=mean*mean/stdDev;
	normalization+=log(erf(mean/sqrt(stdDev)));
	normalization+=0.5*log(stdDev);
        sum=f0*f0+realMean*realMean+imagMean*imagMean;
        sum/=stdDev;
        product=2.0*f0*sqrt(realMean*realMean+imagMean*imagMean)/stdDev;
	prob=product-sum;
	if (2.0*product<50.0) prob+=log(1.0-exp(-2.0*product));
	prob-=normalization;
	prob+=(realMean*realMean+imagMean*imagMean)/stdDev;

	return prob;
}

Real calcAnalyticalLikelihood(Real mean, Real stdDev, Real f0)
{
	Real prob=0;
        Real term1, term2;
        Real normalization;

	if (mean<0) mean=-mean;
        stdDev=2.0*stdDev*stdDev;
        normalization=sqrt(pi)*erf(mean/sqrt(stdDev))*sqrt(stdDev);
        term1=(f0-mean)*(f0-mean)/stdDev;
        term2=(f0+mean)*(f0+mean)/stdDev;
        if (abs(term1)<50.0) prob+=exp(-term1);
        if (abs(term2)<50.0) prob+=exp(-term2);
        prob/=normalization;

	return prob;
}

Real calcAnalyticalLogLikelihood(Real mean, Real stdDev, Real f0)
{
	Real prob=0;
        Real term1, term2;
        Real normalization;

	if (mean<0) mean=-mean;
        stdDev=2.0*stdDev*stdDev;
	normalization=0.5*log(pi);
	//normalization+=log(erf(mean/sqrt(stdDev)));
	normalization+=0.5*log(stdDev);
        term1=(f0-mean)*(f0-mean)/stdDev;
        term2=(f0+mean)*(f0+mean)/stdDev;
	prob=-term1;
	if (4.0*f0*mean/stdDev<50.0) prob+=log(1.0+exp(-4.0*f0*mean/stdDev));
        prob-=normalization;

	return prob;
}

Real calcAnalyticalLogLikelihoodPositiveCorrelation(Real mean, Real stdDev, Real f0)
{
	mean*=sqrt(2.0);
	stdDev*=sqrt(2.0);

	return calcAnalyticalLogLikelihood(mean, stdDev, f0);
}
/*
Real calcAnalyticalLogLikelihoodNegativeCorrelation(Real mean, Real stdDev, Real f0)
{
	//This case will never come up
	Real prob;

	cout <<"mean= "<<mean<<" stdDev= "<<stdDev<<" f0= "<<f0<<endl;

	if (f0*f0<=2.0*mean*mean) return -1000.0;
	prob=log(f0)-f0*f0/(4.0*stdDev*stdDev)-0.5*log(f0*f0-2.0*mean*mean);
	
	cout <<"prob= "<<prob<<endl;

	return prob;
}
*/
Real calcNumericalLogLikelihood(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, Real f0)
{
        Real thetaInc, prob=1.0e-30;
        Real cosTerm, sinTerm;
	Real cutoff=700.0;
	Real normalization;

	if (f0==0) f0=1.0e-8;
        thetaInc=2.0*pi/1000.0;
	normalization=2.0*pi*realStdDev*imagStdDev/thetaInc;
        realStdDev=2.0*realStdDev*realStdDev;
        imagStdDev=2.0*imagStdDev*imagStdDev;
        for (Real theta=0;theta<=2.0*pi;theta+=thetaInc)
        {
                cosTerm=f0*cos(theta)-realMean;
                cosTerm=cosTerm*cosTerm/realStdDev;
                sinTerm=f0*sin(theta)-imagMean;
                sinTerm=sinTerm*sinTerm/imagStdDev;
                if (cosTerm+sinTerm<cutoff) prob+=exp(-cosTerm-sinTerm);
        }
        return log(prob)+log(f0)-log(normalization);
}

void calcEmpiricalLikelihood(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, Real min, Real binSize, Real max, Vector &f)
{
        int bin, nbin, niter=100000;
        Real randReal, randImag, amplitude;

        nbin=int((max-min)/binSize+0.5)+1;
        SafeAlloc(f, nbin, "f");

        for (int i=0;i<niter;i++)
        {
                randReal=realMean+randGaussian(realStdDev);
                randImag=imagMean+randGaussian(imagStdDev);
                amplitude=sqrt(randReal*randReal+randImag*randImag);
                bin=int((amplitude-min)/binSize+0.5);
                if (bin>=0 && bin<nbin) f[bin]+=1.0/(binSize*Real(niter));
        }
	f[0]*=2.0;
}

void calcNumericalLikelihood(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, Real min, Real binSize, Real max, Vector &f)
{
        int nbin;
        Real f0;

        nbin=int((max-min)/binSize+0.5)+1;
        SafeAlloc(f, nbin, "f");

        for (int i=0;i<nbin;i++)
        {
                f0=min+Real(i)*binSize;
                f[i]=calcNumericalLogLikelihood(realMean, imagMean, realStdDev, imagStdDev, f0);
		if (abs(f[i])<300.0) f[i]=exp(f[i]);
		else f[i]=0;
        }
}

void calcAnalyticalLikelihood(Real mean, Real stdDev, Real min, Real binSize, Real max, Vector &f)
{
        int nbin;
        Real f0;
        Real term1, term2;
        Real normalization;

        nbin=int((max-min)/binSize+0.5)+1;
        SafeAlloc(f, nbin, "f");

        stdDev=2.0*stdDev*stdDev;
        normalization=sqrt(pi)*sqrt(stdDev);
        for (int i=0;i<nbin;i++)
        {
                f0=min+Real(i)*binSize;
                term1=(f0-mean)*(f0-mean)/stdDev;
                term2=(-f0-mean)*(-f0-mean)/stdDev;
                if (abs(term1)<50.0) f[i]+=exp(-term1);
                if (abs(term2)<50.0) f[i]+=exp(-term2);
                f[i]/=normalization;
        }
}

void calcAnalyticalLikelihood(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, Real min, Real binSize, Real max, Vector &f)
{
        int nbin;
        Real stdDev, mean;
        Real normalization;
        Real f0, sum, product;
	Real cutoff=700.0;

        nbin=int((max-min)/binSize+0.5)+1;
        SafeAlloc(f, nbin, "f");

        realStdDev=2.0*realStdDev*realStdDev;
        imagStdDev=2.0*imagStdDev*imagStdDev;
        stdDev=sqrt(realStdDev*imagStdDev);
        mean=sqrt(realMean*realMean+imagMean*imagMean);
        //normalization=sqrt(pi)*exp(mean*mean/stdDev)*erf(mean/sqrt(stdDev))*sqrt(stdDev);
        normalization=sqrt(pi)*erf(mean/sqrt(stdDev))*sqrt(stdDev);
        for (int i=0;i<nbin;i++)
        {
                f0=min+Real(i)*binSize;
                sum=f0*f0+realMean*realMean+imagMean*imagMean;
                sum/=stdDev;
                product=2.0*f0*sqrt(realMean*realMean+imagMean*imagMean)/stdDev;                
		if (abs(product-sum)<cutoff) f[i]+=exp(product-sum);
                if (abs(product+sum)<cutoff) f[i]-=exp(-product-sum);
		//f[i]=log(f[i])+(realMean*realMean+imagMean*imagMean)/stdDev-log(pi)-mean*mean/stdDev-log(erf(mean/sqrt(stdDev))-0.5*log(stdDev);
		//if (abs(f[i])<cutoff) f[i]=exp(f[i]);
		//else f[i]=0;
                f[i]/=normalization;
                //f[i]*=exp((realMean*realMean+imagMean*imagMean)/stdDev);
        }
}

Real calcAverageFAnalyticalCentric(Real mean, Real stdDev)
{
        mean=abs(mean);
        stdDev*=sqrt(2.0);
        return mean*erf(mean/stdDev)+stdDev*exp(-mean*mean/(stdDev*stdDev))/sqrt(pi);
}

Real calcAverageFAnalyticalAcentric(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev)
{
        Real a, b, c;
        Real expTerm, ave=0;
        Real term1, term2;
        Real thetaInc;

        thetaInc=2.0*pi/1000.0;
        realStdDev*=sqrt(2.0);
        imagStdDev*=sqrt(2.0);
        for (Real theta=0;theta<=2.0*pi;theta+=thetaInc)
        {
                a=cos(theta)*cos(theta)/(realStdDev*realStdDev)+sin(theta)*sin(theta)/(imagStdDev*imagStdDev);
                b=-2.0*(realMean*cos(theta)/(realStdDev*realStdDev)+imagMean*sin(theta)/(imagStdDev*imagStdDev));
                c=realMean*realMean/(realStdDev*realStdDev)+imagMean*imagMean/(imagStdDev*imagStdDev);
                expTerm=exp(-c+b*b/(4.0*a));
                term1=sqrt(pi)*(2.0*a+b*b)*erfc(b/(2.0*sqrt(a)))/(8.0*a*a*sqrt(a));
                term2=b*exp(-b*b/(4.0*a))/(4.0*a*a);
                ave+=expTerm*(term1-term2);
        }
        ave/=pi*realStdDev*imagStdDev;
        return ave*thetaInc;
}

Real calcAverageFAnalyticalAcentric(Real mean, Real stdDev)
{
        Real term;
        Real k=1.0;
        Real sum=1.0;

        stdDev*=sqrt(2.0);
        term=1.0;

        while (true)
        {
                term*=0.5*mean*mean*(2.0*k+1.0)/(k*k*stdDev*stdDev);
                sum+=term;
                k+=1.0;
                if (term<1.0e-9) break;
	}
        sum*=0.25*sqrt(pi)*stdDev*stdDev*stdDev;
        sum*=2.0*pi*exp(-mean*mean/(stdDev*stdDev));
        sum/=pi*stdDev*stdDev;

	return sum;
}

void calcAverageF(XRayStruct &xray)
{
	int npoint=xray.miller.size();
	Real ratio, mean, stdDev;

	for (int i=0;i<npoint;i++)
	{
		mean=sqrt(xray.realAve[i]*xray.realAve[i]+xray.imagAve[i]*xray.imagAve[i]);
		if (xray.centric[i])
		{
			stdDev=sqrt(xray.realStdDev[i]*xray.realStdDev[i]+xray.imagStdDev[i]*xray.imagStdDev[i]);
			xray.Fave[i]=calcAverageFAnalyticalCentric(mean, stdDev);
		}
		else 
		{
			ratio=xray.realStdDev[i]/xray.imagStdDev[i];
			if (ratio>0.8 && ratio<1.25)
			{
				stdDev=sqrt(xray.realStdDev[i]*xray.imagStdDev[i]);
				xray.Fave[i]=calcAverageFAnalyticalAcentric(mean, stdDev);
			}
			else xray.Fave[i]=calcAverageFAnalyticalAcentric(xray.realAve[i], xray.imagAve[i], xray.realStdDev[i], xray.imagStdDev[i]);
		}
		//cout <<"i= "<<i<<" realAve= "<<xray.realAve[i]<<" imagAve= "<<xray.imagAve[i]<<" realStdDev= "<<xray.realStdDev[i]<<" imagStdDev= "<<xray.imagStdDev[i]<<" xray.Fave= "<<xray.Fave[i]<<endl;
	}
	//endProgram(__LINE__, __FILE__);
}

Real compareNumericalEmpirical(Vector &fNumerical, Vector &fEmpirical, Real min, Real binSize)
{
	int nbin=fNumerical.size();
	Real sum=0, sum2=0, f0;

	for (int i=0;i<nbin;i++)
	{
		sum+=(fNumerical[i]-fEmpirical[i])*(fNumerical[i]-fEmpirical[i]);
		sum2+=fEmpirical[i]*fEmpirical[i];
	}
	cout <<"error= "<<sum/sum2<<endl;
	if (sum>sum2*0.01)
	{
		cout <<"Numerical and empirical methods disagree"<<endl;
		cout <<"f0\tnumerical\tempirical"<<endl;
		for (int i=0;i<nbin;i++)
		{
                	f0=min+Real(i)*binSize;
			cout <<"Comparison\t"<<f0<<"\t"<<fNumerical[i]<<"\t"<<fEmpirical[i]<<endl;
		}
		endProgram(__LINE__, __FILE__);
	}
	return sum/sum2;
}

void compareNumericalEmpirical(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev)
{
	Real max, min;
	Real maxReal, maxImag;
	Real minReal, minImag;
	Real binSize, ratio;
	Vector fEmpirical, fNumerical;

	if (realMean<0) realMean=-realMean;
	if (imagMean<0) imagMean=-imagMean;
	if (realStdDev<0) realStdDev=-realStdDev;
	if (imagStdDev<0) imagStdDev=-imagStdDev;

	maxReal=realMean+realStdDev*4.0;
        maxImag=imagMean+imagStdDev*4.0;
        max=sqrt(maxReal*maxReal+maxImag*maxImag);

	minReal=realMean-realStdDev*4.0;
	minImag=imagMean-imagStdDev*4.0;
	if (minReal<0) minReal=0;
	if (minImag<0) minImag=0;
	min=sqrt(minReal*minReal+minImag*minImag);

        binSize=(max-min)/100.0;

	calcEmpiricalLikelihood(realMean, imagMean, realStdDev, imagStdDev, min, binSize, max, fEmpirical);

	if (imagStdDev!=0) ratio=realStdDev/imagStdDev;
	else ratio=1.0e9;

	//if (ratio>0.99 && ratio<1.01)
	//{
	//	cout <<"ratio= "<<ratio<<endl;
	//	calcAnalyticalLikelihood(realMean, imagMean, realStdDev, imagStdDev, min, binSize, max, fNumerical);
	//}		
	//else if (ratio>0.01 && ratio<100.0)
	if (ratio>0.01 && ratio<100.0)
	{
		calcNumericalLikelihood(realMean, imagMean, realStdDev, imagStdDev, min, binSize, max, fNumerical);
	}
	else if (ratio>100.0)
	{
		calcAnalyticalLikelihood(realMean, realStdDev, min, binSize, max, fNumerical);
	}
	else if (ratio<0.01)
	{
		calcAnalyticalLikelihood(imagMean, imagStdDev, min, binSize, max, fNumerical);
	}
	compareNumericalEmpirical(fNumerical, fEmpirical, min, binSize);
}

Real calcAnalyticalLogLikelihoodPoint(Real realAve, Real imagAve, Real realStdDev, Real imagStdDev, bool centric, Real expF)
{
	bool nonZero;
	Real d=10e-8;
	Real ratio, pointMltf=0;
	Real ave, stdDev;

	//cout <<"realAve= "<<realAve<<" imagAve= "<<imagAve<<" realStdDev= "<<realStdDev<<" imagStdDev= "<<imagStdDev<<" centric= "<<centric<<" expF= "<<expF<<endl;

	//compareNumericalEmpirical(realAve, imagAve, realStdDev, imagStdDev);
	if (abs(realStdDev)<d && abs(imagStdDev)<d) nonZero=false;
	else nonZero=true;
	//cout <<"nonZero= "<<nonZero<<endl;
	if (imagStdDev!=0) ratio=realStdDev/imagStdDev;
	else ratio=1.0e9;

	if (!nonZero) return -1000.0;
	if (!centric)
	{
		if (ratio>0.95 && ratio<1.05) pointMltf=calcLogRice(realAve, imagAve, realStdDev, imagStdDev, expF);
		else pointMltf=calcNumericalLogLikelihood(realAve, imagAve, realStdDev, imagStdDev, expF);
		if (isnan(pointMltf) || isinf(pointMltf))
		{
			pointMltf=calcLogRice(realAve, imagAve, realStdDev, imagStdDev, expF);
		}
	}
	else
	{
		ave=sqrt(realAve*realAve+imagAve*imagAve);
		stdDev=sqrt(realStdDev*realStdDev+imagStdDev*imagStdDev);
		pointMltf=calcAnalyticalLogLikelihood(ave, stdDev, expF);
	}	
	//else
	//{
	//	string errorStr="Cannot accurate calculate log likelihood.";
	//	errorStr+=" real= "+toStr(xray.real[i]);
	//	errorStr+=" imag= "+toStr(xray.imag[i]);
	//	errorStr+=" realStdDev= "+toStr(xray.realStdDev[i]);
	//	errorStr+=" imagStdDev= "+toStr(xray.imagStdDev[i]);
	//	error(errorStr, __LINE__, __FILE__);
	//}
	if (isnan(pointMltf) || isinf(pointMltf))
	{
		endProgram(__LINE__, __FILE__);
	}

	return pointMltf;
}

void compareAnalyticalEmpirical(Real realMean, Real imagMean, Real realStdDev, Real imagStdDev, bool centric, Vector &fs)
{
	int nbin=100;
	Real max, min;
	Real pointMltf;
	Real maxReal, maxImag;
	Real minReal, minImag;
	Real f, binSize;
	Vector mltfs;
	Vector histogram;

	cout <<"realMean= "<<realMean<<" imagMean= "<<imagMean<<" realStdDev= "<<realStdDev<<" imagStdDev= "<<imagStdDev<<" centric= "<<centric<<endl;

	//if (realMean<0) realMean=-realMean;
	//if (imagMean<0) imagMean=-imagMean;
	//if (realStdDev<0) realStdDev=-realStdDev;
	//if (imagStdDev<0) imagStdDev=-imagStdDev;

	maxReal=abs(realMean)+realStdDev*4.0;
        maxImag=abs(imagMean)+imagStdDev*4.0;
        max=sqrt(maxReal*maxReal+maxImag*maxImag);

	minReal=abs(realMean)-realStdDev*4.0;
	minImag=abs(imagMean)-imagStdDev*4.0;
	if (minReal<0) minReal=0;
	if (minImag<0) minImag=0;
	min=sqrt(minReal*minReal+minImag*minImag);

        binSize=(max-min)/Real(nbin);

	calcHistogram(fs, histogram, binSize, min, nbin);
	histogram[0]*=2.0;
	normalize(histogram);
	vectorMultiply(1.0/binSize, histogram);	
	SafeAlloc(mltfs, nbin, "mltfs");

	//cout <<endl;
	for (int i=0;i<nbin;i++)
	{
		f=min+Real(i)*binSize;
		pointMltf=calcAnalyticalLogLikelihoodPoint(realMean, imagMean, realStdDev, imagStdDev, centric, f);
		pointMltf=exp(pointMltf);
		mltfs[i]=pointMltf;
		//cout <<"AnalyticalEmpirical\t"<<i<<"\t"<<pointMltf<<"\t"<<histogram[i]<<endl;
	}
	//cout <<endl;
	compareNumericalEmpirical(mltfs, histogram, min, binSize);
}

void compareAnalyticalEmpirical(XRayStruct &xray, Matrix &fs)
{
	int npoint=xray.miller.size();
	Matrix fsTranspose;

	transpose(fs, fsTranspose);

	for (int i=0;i<npoint;i++)
	{
		cout <<"point= "<<i<<endl;
		compareAnalyticalEmpirical(xray.realAve[i], xray.imagAve[i], xray.realStdDev[i], xray.imagStdDev[i], xray.centric[i], fsTranspose[i]);
	}
}

Real calcAnalyticalLogLikelihood(XRayStruct &xray, XRayStruct &expXRay, Real scale)
{
	int npoint=xray.miller.size();
	Real mltf=0;
	//timeval start, end;
	//vectorMultiply(scale, xray.real);
	//vectorMultiply(scale, xray.imag);
	//vectorMultiply(scale, xray.realAve);
	//vectorMultiply(scale, xray.imagAve);
	//vectorMultiply(scale, xray.realStdDev);
	//vectorMultiply(scale, xray.imagStdDev);
	//vectorMultiply(scale, xray.realRandomStdDev);
	//vectorMultiply(scale, xray.imagRandomStdDev);

	if (scale<0.01 || scale>100.0)
	{	
		cout <<"scale= "<<scale<<endl;
	}
	for (int i=0;i<npoint;i++)
	{
		//cout <<"i= "<<i<<" h= "<<xray.miller[i].Pos[X]<<" k= "<<xray.miller[i].Pos[Y]<<" l= "<<xray.miller[i].Pos[Z]<<" realAve= "<<xray.realAve[i]
		//<<" imagAve= "<<xray.imagAve[i]<<" realStdDev= "<<xray.realStdDev[i]<<" centric= "<<xray.centric[i]<<" Fobs= "<<expXRay.F[i]<<endl;
		//gettimeofday(&start, NULL);
		mltf+=calcAnalyticalLogLikelihoodPoint(xray.realAve[i], xray.imagAve[i], xray.realStdDev[i], xray.imagStdDev[i], xray.centric[i], expXRay.F[i]);
		//gettimeofday(&end, NULL);
		//cout <<"MLTF took "<<calcTimeDiff(start, end)<<endl;
	}
	if (mltf==0) return 1000.0;
	return -mltf/Real(npoint);

}

Real calcAnalyticalLogLikelihood(Vector scale, CalcObsStruct args)
{
	return calcAnalyticalLogLikelihood(args.xray, args.expXRay, scale[0]);
}

Real calcAnalyticalLogLikelihoodOptScale(XRayStruct xray, XRayStruct &expXRay)
{
	Real dScoreCriteria=0.000000001;
	Real maxHours=1.0/2.0;
	Real Scale, mltf;
	Vector scale;
	CalcObsStruct args;
	Real (*func_to_minimize)(Vector, CalcObsStruct)=calcAnalyticalLogLikelihood;


	scaleEachQBin(xray, expXRay);
	addExperimentalErrorToRmsdError(xray, expXRay);
	args.xray=xray;
	args.expXRay=expXRay;
	Scale=calcScaleForSqrDiff(xray.F, expXRay.F);

	SafePushBack(scale, Scale, "scale");
	
	mltf=optimizeCoefficients(scale, args, func_to_minimize, dScoreCriteria, maxHours);
	return mltf;
}

void printStdDev(XRayStruct &xray, XRayStruct &expXRay)
{
	int npoint=xray.miller.size();
	npoint=15;	
	for (int i=0;i<npoint;i++)
	{
		cout <<"i= "<<i<<" realAve= "<<xray.realAve[i]<<" imagAve= "<<xray.imagAve[i]<<" realStdDev= "<<xray.realStdDev[i]<<" imagStdDev= "<<xray.imagStdDev[i]<<" Fave= "<<xray.Fave[i]<<" centric= "<<xray.centric[i]<<" expF= "<<expXRay.F[i]<<endl;
	}
}

Real calcAnalyticalLogLikelihood(XRayStruct &xray, XRayStruct &expXRay)
{
	Real scale, mltf;
	//Vector realStdDevRandom, imagStdDevRandom;
	//timeval start, end;
	//realStdDevRandom=xray.realRandomStdDev;
	//imagStdDevRandom=xray.imagRandomStdDev;

	//gettimeofday(&start, NULL);
	calcAverageF(xray);
	//gettimeofday(&end, NULL);
	//cout <<"calcAverageF took "<<calcTimeDiff(start, end)<<endl;
	//gettimeofday(&start, NULL);
	scaleEachQBinFave(xray, expXRay);
	//gettimeofday(&end, NULL);
	//cout <<"scaleEachQBinFace took "<<calcTimeDiff(start, end)<<endl;
	//gettimeofday(&start, NULL);
	addExperimentalErrorToRmsdError(xray, expXRay);
	//gettimeofday(&end, NULL);
	//cout <<"addExperimentalErrorToRmsdError took "<<calcTimeDiff(start, end)<<endl;
	scale=1.0;
	//gettimeofday(&start, NULL);
	mltf=calcAnalyticalLogLikelihood(xray, expXRay, scale);
	//gettimeofday(&end, NULL);
	//cout <<"calcAnalyticlaLogLikelihood took "<<calcTimeDiff(start, end)<<endl;
	//xray.realRandomStdDev=realStdDevRandom;
	//xray.imagRandomStdDev=imagStdDevRandom;

	return mltf;
}
